@@ -45,6 +45,9 @@
 
 #include "eggshell.h"
 
+#define PCRE2_CODE_UNIT_WIDTH 0
+#include <pcre2.h>
+
 #define URL_MATCH_CURSOR  (GDK_HAND2)
 #define SKEY_MATCH_CURSOR (GDK_HAND2)
 
@@ -158,36 +161,39 @@ typedef struct
 {
 	const char *pattern;
 	TerminalURLFlavour flavor;
-	GRegexCompileFlags flags;
+	guint32 flags;
 } TerminalRegexPattern;
 
 static const TerminalRegexPattern url_regex_patterns[] =
 {
-	{ SCHEME "//(?:" USERPASS "\\@)?" HOST PORT URLPATH, FLAVOR_AS_IS, G_REGEX_CASELESS },
-	{ "(?:www|ftp)" HOSTCHARS_CLASS "*\\." HOST PORT URLPATH , FLAVOR_DEFAULT_TO_HTTP, G_REGEX_CASELESS  },
-	{ "(?:callto:|h323:|sip:)" USERCHARS_CLASS "[" USERCHARS ".]*(?:" PORT "/[a-z0-9]+)?\\@" HOST, FLAVOR_VOIP_CALL, G_REGEX_CASELESS  },
-	{ "(?:mailto:)?" USERCHARS_CLASS "[" USERCHARS ".]*\\@" HOSTCHARS_CLASS "+\\." HOST, FLAVOR_EMAIL, G_REGEX_CASELESS  },
-	{ "news:[[:alnum:]\\Q^_{|}~!\"#$%&'()*+,./;:=?`\\E]+", FLAVOR_AS_IS, G_REGEX_CASELESS  },
+	{ SCHEME "//(?:" USERPASS "\\@)?" HOST PORT URLPATH, FLAVOR_AS_IS, PCRE2_CASELESS },
+	{ "(?:www|ftp)" HOSTCHARS_CLASS "*\\." HOST PORT URLPATH , FLAVOR_DEFAULT_TO_HTTP, PCRE2_CASELESS  },
+	{ "(?:callto:|h323:|sip:)" USERCHARS_CLASS "[" USERCHARS ".]*(?:" PORT "/[a-z0-9]+)?\\@" HOST, FLAVOR_VOIP_CALL, PCRE2_CASELESS  },
+	{ "(?:mailto:)?" USERCHARS_CLASS "[" USERCHARS ".]*\\@" HOSTCHARS_CLASS "+\\." HOST, FLAVOR_EMAIL, PCRE2_CASELESS  },
+	{ "news:[[:alnum:]\\Q^_{|}~!\"#$%&'()*+,./;:=?`\\E]+", FLAVOR_AS_IS, PCRE2_CASELESS  },
 };
 
-static GRegex **url_regexes;
+static VteRegex **url_regexes;
 static TerminalURLFlavour *url_regex_flavors;
 static guint n_url_regexes;
 
+static void terminal_screen_url_match_remove (TerminalScreen *screen);
+
+
 #ifdef ENABLE_SKEY
 static const TerminalRegexPattern skey_regex_patterns[] =
 {
-	{ "s/key [[:digit:]]* [-[:alnum:]]*",         FLAVOR_AS_IS },
-	{ "otp-[a-z0-9]* [[:digit:]]* [-[:alnum:]]*", FLAVOR_AS_IS },
+	{ "s/key [[:digit:]]* [-[:alnum:]]*",         FLAVOR_AS_IS, 0 },
+	{ "otp-[a-z0-9]* [[:digit:]]* [-[:alnum:]]*", FLAVOR_AS_IS, 0 },
 };
 
-static GRegex **skey_regexes;
+static VteRegex **skey_regexes;
 static guint n_skey_regexes;
 
 static void  terminal_screen_skey_match_remove (TerminalScreen            *screen);
 #endif /* ENABLE_SKEY */
 
-G_DEFINE_TYPE (TerminalScreen, terminal_screen, VTE_TYPE_TERMINAL)
+G_DEFINE_TYPE_WITH_PRIVATE (TerminalScreen, terminal_screen, VTE_TYPE_TERMINAL)
 
 static char *
 cwd_of_pid (int pid)
@@ -333,33 +339,22 @@ terminal_screen_init (TerminalScreen *screen)
 		{ "text/x-moz-url",  0, TARGET_MOZ_URL },
 		{ "_NETSCAPE_URL", 0, TARGET_NETSCAPE_URL }
 	};
-	VteTerminal *terminal = VTE_TERMINAL (screen);
 	TerminalScreenPrivate *priv;
 	GtkTargetList *target_list;
 	GtkTargetEntry *targets;
 	int n_targets;
-	guint i;
 
-	priv = screen->priv = G_TYPE_INSTANCE_GET_PRIVATE (screen, TERMINAL_TYPE_SCREEN, TerminalScreenPrivate);
+	priv = screen->priv = terminal_screen_get_instance_private (screen);
 
 	vte_terminal_set_mouse_autohide (VTE_TERMINAL (screen), TRUE);
+#if VTE_CHECK_VERSION (0, 52, 0)
+	vte_terminal_set_bold_is_bright (VTE_TERMINAL (screen), TRUE);
+#endif
 
 	priv->child_pid = -1;
 
 	priv->font_scale = PANGO_SCALE_MEDIUM;
 
-	for (i = 0; i < n_url_regexes; ++i)
-	{
-		TagData *tag_data;
-
-		tag_data = g_slice_new (TagData);
-		tag_data->flavor = url_regex_flavors[i];
-		tag_data->tag = vte_terminal_match_add_gregex (terminal, url_regexes[i], 0);
-		vte_terminal_match_set_cursor_type (terminal, tag_data->tag, URL_MATCH_CURSOR);
-
-		priv->match_tags = g_slist_prepend (priv->match_tags, tag_data);
-	}
-
 	/* Setup DND */
 	target_list = gtk_target_list_new (NULL, 0);
 	gtk_target_list_add_uri_targets (target_list, 0);
@@ -576,20 +571,17 @@ terminal_screen_class_init (TerminalScreenClass *klass)
 	                     G_TYPE_STRV,
 	                     G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
 
-	g_type_class_add_private (object_class, sizeof (TerminalScreenPrivate));
-
 	/* Precompile the regexes */
 	n_url_regexes = G_N_ELEMENTS (url_regex_patterns);
-	url_regexes = g_new0 (GRegex*, n_url_regexes);
+	url_regexes = g_new0 (VteRegex*, n_url_regexes);
 	url_regex_flavors = g_new0 (TerminalURLFlavour, n_url_regexes);
 
 	for (i = 0; i < n_url_regexes; ++i)
 	{
 		GError *error = NULL;
 
-		url_regexes[i] = g_regex_new (url_regex_patterns[i].pattern,
-		                              url_regex_patterns[i].flags | G_REGEX_OPTIMIZE | G_REGEX_MULTILINE,
-		                              0, &error);
+		url_regexes[i] = vte_regex_new_for_match(url_regex_patterns[i].pattern, -1,
+				                         url_regex_patterns[i].flags | PCRE2_MULTILINE, &error);
 		if (error)
 		{
 			g_message ("%s", error->message);
@@ -601,15 +593,14 @@ terminal_screen_class_init (TerminalScreenClass *klass)
 
 #ifdef ENABLE_SKEY
 	n_skey_regexes = G_N_ELEMENTS (skey_regex_patterns);
-	skey_regexes = g_new0 (GRegex*, n_skey_regexes);
+	skey_regexes = g_new0 (VteRegex*, n_skey_regexes);
 
 	for (i = 0; i < n_skey_regexes; ++i)
 	{
 		GError *error = NULL;
 
-		skey_regexes[i] = g_regex_new (skey_regex_patterns[i].pattern,
-		                               G_REGEX_OPTIMIZE | G_REGEX_MULTILINE,
-		                               0, &error);
+		skey_regexes[i] = vte_regex_new_for_match(skey_regex_patterns[i].pattern, -1,
+							  PCRE2_MULTILINE | PCRE2_UTF | PCRE2_NO_UTF_CHECK, &error);
 		if (error)
 		{
 			g_message ("%s", error->message);
@@ -1024,7 +1015,7 @@ terminal_screen_profile_notify_cb (TerminalProfile *profile,
 
 				tag_data = g_slice_new (TagData);
 				tag_data->flavor = FLAVOR_SKEY;
-				tag_data->tag = vte_terminal_match_add_gregex (vte_terminal, skey_regexes[i], 0);
+				tag_data->tag = vte_terminal_match_add_regex (vte_terminal, skey_regexes[i], 0);
 				vte_terminal_match_set_cursor_type (vte_terminal, tag_data->tag, SKEY_MATCH_CURSOR);
 
 				priv->match_tags = g_slist_prepend (priv->match_tags, tag_data);
@@ -1057,6 +1048,29 @@ terminal_screen_profile_notify_cb (TerminalProfile *profile,
 		vte_terminal_set_cursor_shape (vte_terminal,
 		                               terminal_profile_get_property_enum (priv->profile, TERMINAL_PROFILE_CURSOR_SHAPE));
 
+	if (!prop_name || prop_name == I_(TERMINAL_PROFILE_USE_URLS))
+	{
+		if (terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_USE_URLS))
+		{
+			guint i;
+
+			for (i = 0; i < n_url_regexes; ++i)
+			{
+				TagData *tag_data;
+
+				tag_data = g_slice_new (TagData);
+				tag_data->flavor = url_regex_flavors[i];
+				tag_data->tag = vte_terminal_match_add_regex (vte_terminal, url_regexes[i], 0);
+				vte_terminal_match_set_cursor_type (vte_terminal, tag_data->tag, URL_MATCH_CURSOR);
+
+				priv->match_tags = g_slist_prepend (priv->match_tags, tag_data);
+			}
+		}
+		else
+		{
+			terminal_screen_url_match_remove (screen);
+		}
+	}
 	g_object_thaw_notify (object);
 }
 
@@ -1071,6 +1085,7 @@ update_color_scheme (TerminalScreen *screen)
 	const gchar *bg_image_file;
 	double bg_alpha = 1.0;
 	GdkRGBA fg, bg;
+	GdkRGBA *c;
 	guint n_colors;
 	GtkStyleContext *context;
 	GError *error = NULL;
@@ -1079,7 +1094,13 @@ update_color_scheme (TerminalScreen *screen)
 	gtk_style_context_save (context);
 	gtk_style_context_set_state (context, GTK_STATE_FLAG_NORMAL);
 	gtk_style_context_get_color (context, GTK_STATE_FLAG_NORMAL, &fg);
-	gtk_style_context_get_background_color (context, GTK_STATE_FLAG_NORMAL, &bg);
+
+	gtk_style_context_get (context, GTK_STATE_FLAG_NORMAL,
+			       GTK_STYLE_PROPERTY_BACKGROUND_COLOR,
+			       &c, NULL);
+	bg = *c;
+	gdk_rgba_free (c);
+
 	gtk_style_context_restore (context);
 
 	bold_rgba = NULL;
@@ -1138,25 +1159,6 @@ update_color_scheme (TerminalScreen *screen)
 	if (bold_rgba)
 		vte_terminal_set_color_bold (VTE_TERMINAL (screen),
 		                             bold_rgba);
-
-	/* In case of being used in GNOME3 desktop environment.
-	 * code references to gnome-terminal
-	 */
-	if (bg_type == TERMINAL_BACKGROUND_TRANSPARENT)
-	{
-		/* toplevel widget */
-		GtkWidget *toplevel;
-		toplevel = gtk_widget_get_toplevel (GTK_WIDGET (screen));
-		if (toplevel != NULL && gtk_widget_is_toplevel (toplevel) 
-			     && !gtk_widget_get_app_paintable (toplevel))
-		{
-			gtk_widget_set_app_paintable (toplevel, TRUE);
-			/* The opaque region of the toplevel isn't updated until the toplevel is allocated;
-			 * set_app_paintable() doesn't force an allocation, so do that manually.
-			 */
-			gtk_widget_queue_resize (toplevel);
-		}
-	}
 }
 
 void
@@ -1406,7 +1408,7 @@ get_child_environment (TerminalScreen *screen,
 	GHashTableIter iter;
 	GPtrArray *retval;
 	guint i;
-	const char * const *list_schemas;
+	gchar **list_schemas = NULL;
 	gboolean schema_exists;
 
 	window = gtk_widget_get_toplevel (term);
@@ -1440,13 +1442,14 @@ get_child_environment (TerminalScreen *screen,
 	g_hash_table_remove (env_table, "LINES");
 	g_hash_table_remove (env_table, "MATE_DESKTOP_ICON");
 
-	g_hash_table_replace (env_table, g_strdup ("TERM"), g_strdup ("xterm")); /* FIXME configurable later? */
+	g_hash_table_replace (env_table, g_strdup ("TERM"), g_strdup ("xterm-256color")); /* FIXME configurable later? */
 
 	/* FIXME: moving the tab between windows, or the window between displays will make the next two invalid... */
 	g_hash_table_replace (env_table, g_strdup ("WINDOWID"), g_strdup_printf ("%ld", GDK_WINDOW_XID (gtk_widget_get_window (window))));
 	g_hash_table_replace (env_table, g_strdup ("DISPLAY"), g_strdup (gdk_display_get_name (display)));
 
-	list_schemas = g_settings_list_schemas();
+	g_settings_schema_source_list_schemas (g_settings_schema_source_get_default (), TRUE, &list_schemas, NULL);
+
 	schema_exists = FALSE;
 	for (i = 0; list_schemas[i] != NULL; i++) {
 		if (g_strcmp0 (list_schemas[i], CONF_PROXY_SCHEMA) == 0)
@@ -1455,6 +1458,9 @@ get_child_environment (TerminalScreen *screen,
 			break;
 		}
 	}
+
+	g_strfreev (list_schemas);
+
 	if (schema_exists == TRUE) {
 		terminal_util_add_proxy_env (env_table);
 	}
@@ -1506,19 +1512,60 @@ info_bar_response_cb (GtkWidget *info_bar,
 	}
 }
 
+static void handle_error_child (TerminalScreen *screen,
+				GError         *err)
+{
+	GtkWidget *info_bar;
+
+	info_bar = terminal_info_bar_new (GTK_MESSAGE_ERROR,
+	                                  _("_Profile Preferences"), RESPONSE_EDIT_PROFILE,
+	                                  _("_Relaunch"), RESPONSE_RELAUNCH,
+	                                  NULL);
+	terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+	                               _("There was an error creating the child process for this terminal"));
+	terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+	                               "%s", err->message);
+	g_signal_connect (info_bar, "response",
+	                  G_CALLBACK (info_bar_response_cb), screen);
+
+	gtk_box_pack_start (GTK_BOX (terminal_screen_container_get_from_screen (screen)),
+	                    info_bar, FALSE, FALSE, 0);
+	gtk_info_bar_set_default_response (GTK_INFO_BAR (info_bar), GTK_RESPONSE_CANCEL);
+	gtk_widget_show (info_bar);
+
+	g_error_free (err);
+}
+
+static void term_spawn_callback (GtkWidget *terminal,
+				 GPid       pid,
+				 GError    *error,
+				 gpointer   user_data)
+{
+	TerminalScreen *screen = TERMINAL_SCREEN (terminal);
+
+	if (error)
+	{
+		handle_error_child (screen, error);
+		g_error_free (error);
+	}
+	else
+	{
+		TerminalScreenPrivate *priv = screen->priv;
+		priv->child_pid = pid;
+	}
+}
+
 static gboolean
 terminal_screen_launch_child_cb (TerminalScreen *screen)
 {
 	TerminalScreenPrivate *priv = screen->priv;
 	VteTerminal *terminal = VTE_TERMINAL (screen);
-	TerminalProfile *profile;
 	char **env, **argv;
 	char *shell = NULL;
 	GError *err = NULL;
 	const char *working_dir;
 	VtePtyFlags pty_flags = VTE_PTY_DEFAULT;
 	GSpawnFlags spawn_flags = 0;
-	GPid pid;
 
 	priv->launch_child_source_id = 0;
 
@@ -1526,8 +1573,6 @@ terminal_screen_launch_child_cb (TerminalScreen *screen)
 	                       "[screen %p] now launching the child process\n",
 	                       screen);
 
-	profile = priv->profile;
-
 	env = get_child_environment (screen, &shell);
 
 	if (priv->initial_working_directory)
@@ -1535,41 +1580,9 @@ terminal_screen_launch_child_cb (TerminalScreen *screen)
 	else
 		working_dir = g_get_home_dir ();
 
-	if (!terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_LOGIN_SHELL))
-		pty_flags |= VTE_PTY_NO_LASTLOG;
-	if (!terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_UPDATE_RECORDS))
-		pty_flags |= VTE_PTY_NO_UTMP | VTE_PTY_NO_WTMP;
-
-	if (!get_child_command (screen, shell, &spawn_flags, &argv, &err) ||
-	        !vte_terminal_spawn_sync (
-	                terminal,
-	                pty_flags,
-	                working_dir,
-	                argv,
-	                env,
-	                spawn_flags,
-	                NULL, NULL,
-	                &pid,
-	                NULL,
-	                &err))
+	if (!get_child_command (screen, shell, &spawn_flags, &argv, &err))
 	{
-		GtkWidget *info_bar;
-
-		info_bar = terminal_info_bar_new (GTK_MESSAGE_ERROR,
-		                                  _("_Profile Preferences"), RESPONSE_EDIT_PROFILE,
-		                                  _("_Relaunch"), RESPONSE_RELAUNCH,
-		                                  NULL);
-		terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
-		                               _("There was an error creating the child process for this terminal"));
-		terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
-		                               "%s", err->message);
-		g_signal_connect (info_bar, "response",
-		                  G_CALLBACK (info_bar_response_cb), screen);
-
-		gtk_box_pack_start (GTK_BOX (terminal_screen_container_get_from_screen (screen)),
-		                    info_bar, FALSE, FALSE, 0);
-		gtk_info_bar_set_default_response (GTK_INFO_BAR (info_bar), GTK_RESPONSE_CANCEL);
-		gtk_widget_show (info_bar);
+		handle_error_child (screen, err);
 
 		g_error_free (err);
 		g_strfreev (env);
@@ -1578,7 +1591,19 @@ terminal_screen_launch_child_cb (TerminalScreen *screen)
 		return FALSE;
 	}
 
-	priv->child_pid = pid;
+        vte_terminal_spawn_async (terminal,
+				  pty_flags,
+				  working_dir,
+				  argv,
+				  env,
+				  spawn_flags,
+				  NULL,
+				  NULL,
+				  NULL,
+				  -1,
+				  NULL,
+				  (VteTerminalSpawnAsyncCallback) term_spawn_callback,
+				  NULL);
 
 	g_free (shell);
 	g_strfreev (argv);
@@ -1925,13 +1950,17 @@ terminal_screen_child_exited (VteTerminal *terminal, int status)
 	switch (action)
 	{
 	case TERMINAL_EXIT_CLOSE:
-		g_signal_emit (screen, signals[CLOSE_SCREEN], 0);
+		if ((status != 9) || (priv->override_command != NULL))
+			g_signal_emit (screen, signals[CLOSE_SCREEN], 0);
 		break;
 	case TERMINAL_EXIT_RESTART:
 		terminal_screen_launch_child_on_idle (screen);
 		break;
 	case TERMINAL_EXIT_HOLD:
 	{
+		if ((status == 9) && (priv->override_command == NULL))
+			break;
+
 		GtkWidget *info_bar;
 
 		info_bar = terminal_info_bar_new (GTK_MESSAGE_INFO,
@@ -2298,6 +2327,30 @@ terminal_screen_skey_match_remove (TerminalScreen *screen)
 }
 #endif /* ENABLE_SKEY */
 
+static void
+terminal_screen_url_match_remove (TerminalScreen *screen)
+{
+	TerminalScreenPrivate *priv = screen->priv;
+	GSList *l, *next;
+
+	l = priv->match_tags;
+	while (l != NULL)
+	{
+		TagData *tag_data = (TagData *) l->data;
+
+		next = l->next;
+#ifdef ENABLE_SKEY
+		if (tag_data->flavor != FLAVOR_SKEY)
+#endif
+		{
+			vte_terminal_match_remove (VTE_TERMINAL (screen), tag_data->tag);
+			priv->match_tags = g_slist_delete_link (priv->match_tags, l);
+		}
+
+		l = next;
+	}
+}
+
 static char*
 terminal_screen_check_match (TerminalScreen *screen,
                              GdkEvent  *event,
